/* -*- linux-c -*- */

/* gtkcomboboxgrid.c - GtkComboBoxGrid widget implementation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkcomboboxgrid.h"

#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkobject.h>
#include <gtk/gtktable.h>

enum {
	SELECT,
	LAST_SIGNAL
};

static void   gtk_combo_box_grid_destroy                   ( GtkObject *object );

static void   gtk_combo_box_grid_class_init                ( GtkComboBoxGridClass *class );

static void   gtk_combo_box_grid_calculate_dimensions      ( GtkComboBoxGrid *cboxgrid );
static gint   gtk_combo_box_grid_get_largest_height        ( GtkComboBoxGrid *cboxgrid );
static gint   gtk_combo_box_grid_get_largest_width         ( GtkComboBoxGrid *cboxgrid );

static GtkComboBoxGridItem *gtk_combo_box_grid_item_create ( GtkComboBoxGrid *cboxgrid,
							     GtkWidget       *widget );
static void   gtk_combo_box_grid_item_delete               ( GtkComboBoxGrid *cboxgrid,
							     GtkComboBoxGridItem *item );
static void   gtk_combo_box_grid_item_delete_from_list     ( GtkComboBoxGrid *cboxgrid,
							     GtkComboBoxGridItem *item );
static void   _gtk_combo_box_grid_item_select              ( GtkComboBoxGrid     *cboxgrid,
							     GtkComboBoxGridItem *item,
							     gint mode );
static void   gtk_combo_box_grid_clicked                   ( GtkWidget *button,
							     GtkComboBoxGrid *cboxgrid );
static void   gtk_combo_box_grid_emit_change               ( GtkWidget *widget,
							     GtkComboBoxGrid *cboxgrid );
static void   gtk_combo_box_grid_table_setup               ( GtkComboBoxGrid *cboxgrid );
static void   gtk_combo_box_grid_table_clear               ( GtkComboBoxGrid *cboxgrid );
static void   gtk_combo_box_grid_table_rebuild             ( GtkComboBoxGrid *cboxgrid );


static gint                   gtk_combo_box_grid_signals[LAST_SIGNAL] = { 0 };
static GtkComboBoxClass      *gtk_combo_box_grid_parent_class = NULL;

/* object class functions */

static void
gtk_combo_box_grid_destroy ( GtkObject  *object )
{
	GtkComboBoxGrid *cboxgrid = GTK_COMBO_BOX_GRID ( object );

	g_return_if_fail ( object != NULL );
	
	if ( cboxgrid->box_items ) {
		GList   *i;

		gtk_combo_box_grid_table_clear ( cboxgrid );

		for ( i = cboxgrid->box_items ; i != NULL ; i = i->next )
			if ( i->data ) {
				gtk_combo_box_grid_item_delete_from_list (
					cboxgrid,
					(GtkComboBoxGridItem *)i->data );
				gtk_combo_box_grid_item_delete ( 
					cboxgrid, 
					(GtkComboBoxGridItem *)i->data );
				i = cboxgrid->box_items;
			}

		g_list_free ( cboxgrid->box_items );
		cboxgrid->box_items = NULL;
	}

	if ( cboxgrid->preview_button ) {
		gtk_widget_unparent ( cboxgrid->preview_button );
		gtk_widget_unref ( cboxgrid->preview_button );
		cboxgrid->preview_button = NULL;
	}

	GTK_OBJECT_CLASS ( gtk_combo_box_grid_parent_class )->destroy (
		object );
}

/* class functions */

static void
gtk_combo_box_grid_class_init ( GtkComboBoxGridClass *class )
{
	GtkObjectClass    *object_class;
	GtkContainerClass *container_class;
	
	g_return_if_fail ( class != NULL );

	g_warning ( "GtkComboBoxGrid: this widget is far from stable\n" );

	object_class = (GtkObjectClass *)class;
	container_class = (GtkContainerClass *)class;

	object_class->destroy = gtk_combo_box_grid_destroy;

	gtk_combo_box_grid_parent_class = g_type_class_peek_parent ( class );

	gtk_combo_box_grid_signals[SELECT] =
		gtk_signal_new ( 
			"select",
			GTK_RUN_LAST,
			gtk_combo_box_grid_get_type ( ),
			GTK_SIGNAL_OFFSET ( GtkComboBoxGridClass, select ),
			gtk_marshal_NONE__POINTER,
			GTK_TYPE_NONE, 1, GTK_TYPE_INT );
	
	gtk_object_class_add_signals ( object_class,
				       gtk_combo_box_grid_signals,
				       LAST_SIGNAL );
}

/* utility functions */

static void
gtk_combo_box_grid_calculate_dimensions ( GtkComboBoxGrid *cboxgrid )
{
	double   n, a, b;
	double   x, y;
	double   sx;

	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	
	/**
	 * Thanks go to my maths teacher for helping me figuring
	 * the following algorithm out.
	 */

	/*
	 * FIXME: still have to fine tune this
	 */
	
	n = (double)cboxgrid->number_of_elements - 1;
	a = (double)gtk_combo_box_grid_get_largest_width ( cboxgrid );
	b = (double)gtk_combo_box_grid_get_largest_height ( cboxgrid );

	y = 0.0;
	sx = 1;

	for ( x = 1.0 ; x <= n ; x++ ) {
		double tmp;

		y = (n / x) * a;
		
		if ( y != ((double)((int)y)) ) continue;
		
		tmp = abs ( (y - (x * b)) );
		
		if ( tmp < abs ( ((n / sx) * a) - (sx * b) ) ) sx = x;
	}
	
	cboxgrid->rows = sx;
	cboxgrid->cols = n / sx;

	if ( cboxgrid->rows < 1 )
		cboxgrid->rows = 1;
	if ( cboxgrid->cols < 1 )
		cboxgrid->cols = 1;

	gtk_table_resize ( GTK_TABLE ( cboxgrid->table ),
			   cboxgrid->cols,
			   cboxgrid->rows );
}

static gint
gtk_combo_box_grid_get_largest_height ( GtkComboBoxGrid *cboxgrid )
{
	gint     largest = 0;
	GList   *i;

	g_return_val_if_fail ( cboxgrid != NULL, -1 );
	g_return_val_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ), -1 );
	g_return_val_if_fail ( cboxgrid->box_items != NULL, -1 );

	for ( i = cboxgrid->box_items ; i != NULL ; i = i->next ) {
		GtkComboBoxGridItem  *tmp =
			(GtkComboBoxGridItem *)i->data;

		if ( !tmp ) continue;
		
		if ( tmp->widget->allocation.height > largest )
			largest = tmp->widget->allocation.height;
	}

	return largest;
}

static gint
gtk_combo_box_grid_get_largest_width ( GtkComboBoxGrid *cboxgrid )
{
	gint     largest = 0;
	GList   *i;

	g_return_val_if_fail ( cboxgrid != NULL, -1 );
	g_return_val_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ), -1 );
	g_return_val_if_fail ( cboxgrid->box_items != NULL, -1 );
	
	for ( i = cboxgrid->box_items ; i != NULL ; i = i->next ) {
		GtkComboBoxGridItem *tmp =
			(GtkComboBoxGridItem *)i->data;
		
		if ( !tmp ) continue;
		
		if ( tmp->widget->allocation.width > largest )
			largest = tmp->widget->allocation.width;
	}

	return largest;
}

static GtkComboBoxGridItem *
gtk_combo_box_grid_item_create ( GtkComboBoxGrid *cboxgrid,
				 GtkWidget *widget )
{
	GtkComboBoxGridItem   *item;

	g_return_val_if_fail ( cboxgrid != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ), NULL );
	g_return_val_if_fail ( widget != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_WIDGET ( widget ), NULL );

	item = g_new ( GtkComboBoxGridItem, 1 );
	item->index = -1;
	item->widget = widget;
	item->button = gtk_button_new ( );

	gtk_button_set_relief ( GTK_BUTTON ( item->button ),
				GTK_RELIEF_NONE );
	gtk_container_add ( GTK_CONTAINER ( item->button ),
			    item->widget );
	gtk_signal_connect ( 
		GTK_OBJECT ( item->button ), "clicked",
		GTK_SIGNAL_FUNC ( gtk_combo_box_grid_clicked ),
		cboxgrid );
	gtk_object_set_user_data ( GTK_OBJECT ( item->button ),
				   item );
	gtk_widget_ref ( item->widget );
	gtk_widget_ref ( item->button );
	gtk_widget_show_all ( item->button );	

	cboxgrid->box_items = g_list_append ( cboxgrid->box_items,
					      item );
	cboxgrid->number_of_elements++;

	return item;
}

static void
gtk_combo_box_grid_item_delete ( GtkComboBoxGrid *cboxgrid,
				 GtkComboBoxGridItem *item )
{
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( item != NULL );

	if ( item->widget ) {
		gtk_widget_unref ( item->widget );
		item->widget = NULL;
	}
	
	if ( item->button ) {
		gtk_widget_unref ( item->button );
		item->button = NULL;
	}	
	
	g_free ( item );
	item = NULL;
}

static void
gtk_combo_box_grid_item_delete_from_list ( GtkComboBoxGrid *cboxgrid,
					   GtkComboBoxGridItem *item )
{
	GList  *i;
	GList  *link = NULL;

	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( item != NULL );
	
	gtk_widget_hide ( item->button );
	gtk_widget_unmap ( item->button );
	
	gtk_widget_hide ( item->widget );
	gtk_widget_unmap ( item->widget );

	for ( i = cboxgrid->box_items ; i != NULL ; i = i->next ) {
		if ( i->data == item ) {
			link = i;
			break;
		}
	}

	g_return_if_fail ( link != NULL );

	cboxgrid->box_items = g_list_remove_link ( cboxgrid->box_items,
						   link );
	cboxgrid->number_of_elements--;
}

static void
_gtk_combo_box_grid_item_select ( GtkComboBoxGrid     *cboxgrid,
				  GtkComboBoxGridItem *item,
				  gint mode )
{
	GtkWidget  *tmp;
	
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( item != NULL );
	
	if ( item->widget == cboxgrid->selected->widget )
		return;

	if ( mode )
		gtk_combo_box_grid_table_clear ( cboxgrid );

	/* NOTE: this is a dirty hack :) The button switcher :)
	 *
	 * I wrote this, because I was frustrated I couldn't find
	 * another way to get the buttons right :)
	 */

	gtk_container_remove ( GTK_CONTAINER ( item->button ),
			       item->widget );
	gtk_container_remove ( GTK_CONTAINER ( cboxgrid->selected->button ),
			       cboxgrid->selected->widget );

	gtk_container_add ( GTK_CONTAINER ( item->button ),
			    cboxgrid->selected->widget );
	gtk_container_add ( GTK_CONTAINER ( cboxgrid->selected->button ),
			    item->widget );

	
	tmp = item->button;
	item->button = cboxgrid->selected->button;
	cboxgrid->selected->button = tmp;
	
	cboxgrid->selected = item;

	gtk_widget_show_all ( item->button );
	gtk_widget_show_all ( cboxgrid->selected->button );
	
	gtk_combo_box_grid_table_rebuild ( cboxgrid );
}

static void
gtk_combo_box_grid_clicked ( GtkWidget *button, GtkComboBoxGrid *cboxgrid )
{
	GtkComboBoxGridItem     *item;

	g_return_if_fail ( button != NULL );
	g_return_if_fail ( GTK_IS_BUTTON ( button ) );
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );

	item = gtk_combo_box_grid_item_get_by_widget ( 
		cboxgrid, GTK_BIN ( button )->child );
	
	g_return_if_fail ( item != NULL );

	gtk_combo_box_popup_hide ( GTK_COMBO_BOX ( cboxgrid ) );

	gtk_combo_box_grid_item_select ( cboxgrid, item );
	gtk_combo_box_grid_emit_change ( NULL, cboxgrid );
}

static void
gtk_combo_box_grid_emit_change ( GtkWidget *widget,
				 GtkComboBoxGrid *cboxgrid )
{
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( cboxgrid->selected != NULL );

	gtk_signal_emit ( GTK_OBJECT ( cboxgrid ),
			  gtk_combo_box_grid_signals[SELECT],
			  cboxgrid->selected );
}

static void
gtk_combo_box_grid_table_setup ( GtkComboBoxGrid *cboxgrid )
{
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( cboxgrid->table == NULL );

	cboxgrid->table = gtk_table_new ( 1, 1, FALSE );
	gtk_widget_ref ( cboxgrid->table );
	gtk_widget_show ( cboxgrid->table );
}

static void
gtk_combo_box_grid_table_clear ( GtkComboBoxGrid *cboxgrid )
{
	GList   *i;

	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );

	for ( i = cboxgrid->box_items ; i != NULL ; i = i->next ) {
		GtkComboBoxGridItem   *item =
			(GtkComboBoxGridItem *)i->data;

		if ( item->widget != cboxgrid->selected->widget )
			gtk_container_remove ( 
				GTK_CONTAINER ( cboxgrid->table ),
				item->button );
	}
}

static void
gtk_combo_box_grid_table_rebuild ( GtkComboBoxGrid *cboxgrid )
{
	int          row, col;
	GList       *i;
	
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( cboxgrid->table != NULL );

	gtk_combo_box_grid_calculate_dimensions ( cboxgrid );

	row = col = 0;
	i = cboxgrid->box_items;

	for ( col = 0 ; col < cboxgrid->cols && i != NULL ; col++ ) {
		for ( row = 0 ; row < cboxgrid->rows && i != NULL ; row++ ) {
			GtkComboBoxGridItem  *item =
				(GtkComboBoxGridItem *)i->data;

			if ( item->widget != cboxgrid->selected->widget ) {
				gtk_table_attach ( 
					GTK_TABLE ( cboxgrid->table ),
					item->button,
					col, col + 1, row, row + 1,
					GTK_FILL, GTK_FILL, 1, 1 );
				gtk_widget_show_all ( item->button );
			} else
				row--;
			
			i = i->next;
		}
	}
}

/* implementation of the functions from the header file */

GtkComboBoxGridItem *
gtk_combo_box_grid_item_add ( GtkComboBoxGrid *cboxgrid,
			      GtkWidget *widget )
{
	GtkComboBoxGridItem   *item;

	g_return_val_if_fail ( cboxgrid != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_COMBO_BOX ( cboxgrid ), NULL );
	g_return_val_if_fail ( widget != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_WIDGET ( widget ), NULL );
	
	gtk_combo_box_grid_table_clear ( cboxgrid );

	item = gtk_combo_box_grid_item_create ( cboxgrid, widget );

	gtk_combo_box_grid_table_rebuild ( cboxgrid );

	return item;
}

void
gtk_combo_box_grid_construct ( GtkComboBoxGrid *cboxgrid,
			       GtkWidget       *widget )
{
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( widget != NULL );
	g_return_if_fail ( GTK_IS_WIDGET ( widget ) );

	cboxgrid->cols = cboxgrid->rows = cboxgrid->number_of_elements = 0;
	cboxgrid->box_items = NULL;

	gtk_combo_box_grid_table_setup ( cboxgrid );

	cboxgrid->selected = gtk_combo_box_grid_item_create ( cboxgrid,
							      widget );

	gtk_combo_box_grid_table_clear ( cboxgrid );

	cboxgrid->preview_button = cboxgrid->selected->button;
	gtk_widget_ref ( cboxgrid->preview_button );
	gtk_signal_connect (
		GTK_OBJECT ( cboxgrid->preview_button ), "clicked",
		GTK_SIGNAL_FUNC ( gtk_combo_box_grid_emit_change ),
		cboxgrid );
	gtk_signal_disconnect_by_func ( 
		GTK_OBJECT ( cboxgrid->selected->button ),
		GTK_SIGNAL_FUNC ( gtk_combo_box_grid_clicked ),
		cboxgrid );
	gtk_widget_show_all ( cboxgrid->preview_button );

	gtk_combo_box_grid_table_rebuild ( cboxgrid );
	
	gtk_combo_box_construct ( GTK_COMBO_BOX ( cboxgrid ),
				  cboxgrid->preview_button,
				  cboxgrid->table );
}

GtkComboBoxGridItem *
gtk_combo_box_grid_item_get_by_widget ( GtkComboBoxGrid *cboxgrid,
					GtkWidget       *widget )
{
	GList                  *i;
	GtkComboBoxGridItem    *result = NULL;

	g_return_val_if_fail ( cboxgrid != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ), NULL );
	g_return_val_if_fail ( widget != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_WIDGET ( widget ), NULL );

	for ( i = cboxgrid->box_items ; i != NULL ; i = i->next ) {
		GtkComboBoxGridItem   *tmp =
			(GtkComboBoxGridItem *)i->data;

		if ( !tmp ) continue;

		if ( tmp->widget == widget ) {
			result = tmp;
			break;
		}
	}

	return result;
}

GtkType
gtk_combo_box_grid_get_type ( void )
{
	static GtkType  type = 0;

	if ( !type ) {
		static const GTypeInfo info = {
			sizeof ( GtkComboBoxGridClass ),
			NULL,
			NULL,
			(GClassInitFunc) gtk_combo_box_grid_class_init,
			NULL,
			NULL,
			sizeof ( GtkComboBoxGrid ),
			0,
			NULL
		};

		type = g_type_register_static ( gtk_combo_box_get_type ( ),
						"GtkComboBoxGrid",
						&info );
	}

	return type;
}

GtkWidget *
gtk_combo_box_grid_new ( GtkWidget *widget )
{
	GtkComboBoxGrid   *cboxgrid;

	g_return_val_if_fail ( widget != NULL, NULL );
	g_return_val_if_fail ( GTK_IS_WIDGET ( widget ), NULL );

	cboxgrid = gtk_type_new ( gtk_combo_box_grid_get_type ( ) );
	gtk_combo_box_grid_construct ( cboxgrid, widget );

	return GTK_WIDGET ( cboxgrid );
}

void
gtk_combo_box_grid_item_remove_by_widget ( GtkComboBoxGrid *cboxgrid,
					   GtkWidget *widget )
{
	GList   *i;

	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( widget != NULL );
	g_return_if_fail ( GTK_IS_WIDGET ( widget ) );

	i = cboxgrid->box_items;
	
	for ( ; i != NULL ; i = i->next ) {
		GtkComboBoxGridItem   *ii =
			(GtkComboBoxGridItem *)i->data;

		if ( ii->widget == widget )
			gtk_combo_box_grid_item_remove ( cboxgrid,
							 ii );
	}
}

void
gtk_combo_box_grid_item_remove ( GtkComboBoxGrid     *cboxgrid,
				 GtkComboBoxGridItem *item )
{
	GtkComboBoxGridItem    *new_selected = NULL;

	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( item != NULL );

	gtk_combo_box_grid_table_clear ( cboxgrid );

	if ( item->widget == cboxgrid->selected->widget ) {
		GList *i = cboxgrid->box_items;

		/* the widget we want to remove is selected,
		 * so we'll select another one here
		 */

		for ( ; i != NULL ; i = i->next ) {
			GtkComboBoxGridItem *ii = 
				(GtkComboBoxGridItem *)i->data;

			if ( ii->widget != item->widget ) {
				new_selected = ii;
				break;
			}
		}
	}
	
	gtk_combo_box_grid_item_delete_from_list ( cboxgrid, item );

	/* select_item will do the table rebuilding */
	
	if ( new_selected )
		_gtk_combo_box_grid_item_select ( cboxgrid, new_selected, 0  );
	else
		gtk_combo_box_grid_table_rebuild ( cboxgrid );
	
	gtk_combo_box_grid_item_delete ( cboxgrid, item );
}

void gtk_combo_box_grid_item_select ( GtkComboBoxGrid *cboxgrid,
				      GtkComboBoxGridItem *item )
{
	g_return_if_fail ( cboxgrid != NULL );
	g_return_if_fail ( GTK_IS_COMBO_BOX_GRID ( cboxgrid ) );
	g_return_if_fail ( item != NULL );
	
	_gtk_combo_box_grid_item_select ( cboxgrid, item, 1 );
}
